/*
 * SpanDSP - a series of DSP components for telephony
 *
 * t38.h - An implementation of T.38, less the packet exchange part
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2005 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: t38.h,v 1.4 2005/10/13 17:22:51 steveu Exp $
 */

/*! \file */

#if !defined(_T38_H_)
#define _T38_H_

/*! \page T38_page T.38 real time FAX over IP handling
\section T38_page_sec_1 What does it do?
The T.38 real time FAX over IP module.....

\section T38_page_sec_2 How does it work?
*/

enum
{
    T38_IND_NO_SIGNAL = 0,
    T38_IND_CNG,
    T38_IND_CED,
    T38_IND_V21_PREAMBLE,
    T38_IND_V27TER_2400_TRAINING,
    T38_IND_V27TER_4800_TRAINING,
    T38_IND_V29_7200_TRAINING,
    T38_IND_V29_9600_TRAINING,
    T38_IND_V17_7200_SHORT_TRAINING,
    T38_IND_V17_7200_LONG_TRAINING,
    T38_IND_V17_9600_SHORT_TRAINING,
    T38_IND_V17_9600_LONG_TRAINING,
    T38_IND_V17_12000_SHORT_TRAINING,
    T38_IND_V17_12000_LONG_TRAINING,
    T38_IND_V17_14400_SHORT_TRAINING,
    T38_IND_V17_14400_LONG_TRAINING,
    T38_IND_V8_ANSAM,
    T38_IND_V8_SIGNAL,
    T38_IND_V34_CNTL_CHANNEL_1200,
    T38_IND_V34_PRI_CHANNEL,
    T38_IND_V34_CC_RETRAIN,
    T38_IND_V33_12000_TRAINING,
    T38_IND_V33_14400_TRAINING
} t30_indicator_types_e;

enum
{
    T38_DATA_V21 = 0,
    T38_DATA_V27TER_2400,
    T38_DATA_V27TER_4800,
    T38_DATA_V29_7200,
    T38_DATA_V29_9600,
    T38_DATA_V17_7200,
    T38_DATA_V17_9600,
    T38_DATA_V17_12000,
    T38_DATA_V17_14400,
    T38_DATA_V8,
    T38_DATA_V34_PRI_RATE,
    T38_DATA_V34_CC_1200,
    T38_DATA_V34_PRI_CH,
    T38_DATA_V33_12000,
    T38_DATA_V33_14400
} t38_data_types_e;

enum
{
    T38_FIELD_HDLC_DATA = 0,
    T38_FIELD_HDLC_SIG_END,
    T38_FIELD_HDLC_FCS_OK,
    T38_FIELD_HDLC_FCS_BAD,
    T38_FIELD_HDLC_FCS_OK_SIG_END,
    T38_FIELD_HDLC_FCS_BAD_SIG_END,
    T38_FIELD_T4_NON_ECM_DATA,
    T38_FIELD_T4_NON_ECM_SIG_END,
    T38_FIELD_CM_MESSAGE,
    T38_FIELD_JM_MESSAGE,
    T38_FIELD_CI_MESSAGE,
    T38_FIELD_V34RATE
} t38_field_types_e;

enum
{
    T38_TYPE_OF_MSG_T30_INDICATOR = 0,
    T38_TYPE_OF_MSG_T30_DATA
};

typedef struct
{
    int                 numocts;
    const uint8_t       *data;
} asn1_dyn_oct_str_t;

typedef struct
{
    uint8_t             field_data_present;
    unsigned int        field_type;
    asn1_dyn_oct_str_t  field_data;
} data_field_element_t;

typedef struct t38_state_s t38_state_t;

typedef int (t38_tx_packet_handler_t)(t38_state_t *s, void *user_data, const uint8_t *buf, int len);

enum
{
    T38_TRANSPORT_UDPTL,
    T38_TRANSPORT_RTP,
    T38_TRANSPORT_TCP
};

#define T38_TX_BUF_LEN  16384

#include <sys/time.h>

struct t38_state_s
{
    int gateway;

    /*! Method 1: Local generation of TCF (required for use with TCP).
        Method 2: Transfer of TCF is required for use with UDP (UDPTL or RTP).
        Method 2 is not recommended for use with TCP. */
    int data_rate_management_method;
    
    /*! The emitting gateway may indicate a preference for either UDP/UDPTL, or
        UDP/RTP, or TCP for transport of T.38 IFP Packets. The receiving device
        selects the transport protocol. */
    int data_transport_protocol;

    /*! Indicates the capability to remove and insert fill bits in Phase C, non-ECM
        data to reduce bandwidth in the packet network. Optional. See Note. */
    int fill_bit_removal;

    /*! Indicates the ability to convert to/from MMR from/to the line format to
        improve the compression of the data, and reduce the bandwidth, in the
        packet network. Optional. See Note. */
    int mmr_transcoding;

    /*! Indicates the ability to convert to/from JBIG to reduce bandwidth. Optional.
        See Note. */
    int jbig_transcoding;

    /*! For UDP (UDPTL or RTP) modes, this option indicates the maximum
        number of octets that can be stored on the remote device before an overflow
        condition occurs. It is the responsibility of the transmitting application to
        limit the transfer rate to prevent an overflow. The negotiated data rate
        should be used to determine the rate at which data is being removed from
        the buffer. */
    int max_buffer_size;

    /*! This option indicates the maximum size of a UDPTL packet or the
        maximum size of the payload within an RTP packet that can be accepted by
        the remote device. */
    int max_datagram_size;

    /*! This option indicates the error correction scheme used in a UDPTL packet. */
    int error_correction_scheme;
    int error_correction_entries;
    int error_correction_span;

    /*! This is the version number of ITU-T Rec. T.38. New versions shall be
        compatible with previous versions. */
    int t38_version;

    /*! NOTE  Bandwidth reduction shall only be done on suitable Phase C data, i.e., MH, MR and  in the case
        of transcoding to JBIG  MMR. MMR and JBIG require reliable data transport such as that provided by
        TCP. When transcoding is selected, it shall be applied to every suitable page in a call. */

    char hdlc_buf[256];
    int hdlc_len;
    char hdlc_buf2[256];
    int hdlc_len2;
    char hdlc_buf3[256];
    int hdlc_len3;
    int hdlc_ptr;
    int timed_step;

    uint8_t rx_data[2048];
    int rx_data_bytes;

    uint8_t tx_data[T38_TX_BUF_LEN];
    int tx_in_bytes;
    int tx_data_bytes;

    /*! \brief The location of the most recent EOL marker in the non-ECM data buffer */
    int tx_latest_eol_byte;
    unsigned int bit_stream;
    int bit_no;
    int current_byte;

    int current_rx_indicator;
    int next_rx_indicator;

    int current_tx_indicator;
    int next_tx_indicator;
    int current_tx_data;

    int tx_seq_no;
    int rx_seq_no;
    int rx_expected_seq_no;

    int rx_flag_count;

    /*! \brief The current bit rate for the fast modem. */
    int bit_rate;
    /*! \brief TRUE is a carrier is presnt. Otherwise FALSE. */
    int rx_signal_present;

    /*! \brief A tone generator context used to generate supervisory tones during
               FAX handling. */
    tone_gen_state_t tone_gen;
    /*! \brief An HDLC context used when receiving HDLC over V.21 messages. */
    hdlc_rx_state_t hdlcrx;

    /*! \brief HDLC data used when transmitting HDLC over V.21 messages. */
    int hdlctx_num_bits;
    int hdlctx_idle_byte;
    int hdlctx_len;
    uint8_t hdlctx_buffer[HDLC_MAXFRAME_LEN + 2];
    int hdlctx_pos;
    int hdlctx_byte;
    int hdlctx_bits;
    int hdlctx_ones;

    int hdlc_underflow_reported;
    /*! \brief A V.21 FSK modem context used when transmitting HDLC over V.21
               messages. */
    fsk_tx_state_t v21tx;
    /*! \brief A V.21 FSK modem context used when receiving HDLC over V.21
               messages. */
    fsk_rx_state_t v21rx;

#if defined(ENABLE_V17)
    /*! \brief A V.17 modem context used when sending FAXes at 7200bps, 9600bps
               12000bps or 14400bps*/
    v17_tx_state_t v17tx;
    /*! \brief A V.29 modem context used when receiving FAXes at 7200bps, 9600bps
               12000bps or 14400bps*/
    v17_rx_state_t v17rx;
#endif

    /*! \brief A V.29 modem context used when sending FAXes at 7200bps or
               9600bps */
    v29_tx_state_t v29tx;
    /*! \brief A V.29 modem context used when receiving FAXes at 7200bps or
               9600bps */
    v29_rx_state_t v29rx;

    /*! \brief A V.27ter modem context used when sending FAXes at 2400bps or
               4800bps */
    v27ter_tx_state_t v27ter_tx;
    /*! \brief A V.27ter modem context used when receiving FAXes at 2400bps or
               4800bps */
    v27ter_rx_state_t v27ter_rx;

    int v21_rx_active;
    int v17_rx_active;
    int v27ter_rx_active;
    int v29_rx_active;
    int since_last_tx_packet;
    int silence;
    int short_train;
    int data_final;

    t30_state_t t30_state;
    t38_tx_packet_handler_t *tx_packet_handler;
    void *tx_packet_user_data;

    /*! \brief TRUE if we need to corrupt the HDLC frame in progress, so the receiver cannot
               interpret it. */
    int corrupt_frame;
    
    int current_rx_type;
    int current_tx_type;

    int octets_per_non_ecm_packet;

    struct timeval next_send_time;
    struct timeval last_time;
    
    logging_state_t logging;
};

#ifdef __cplusplus
extern "C" {
#endif

uint8_t bit_reverse(uint8_t data);

char *t38_indicator(int indicator);

char *t38_data_type(int data_type);

char *t38_field_type(int field_type);

void t38_call_event(t38_state_t *s, int event);

int t38_encode_data(t38_state_t *s, uint8_t buf[], int data_type, int field_type, const uint8_t *msg, int msglen);

int t38_encode_indicator(t38_state_t *s, uint8_t buf[], int indicator);

int t38_gateway_process_rx_indicator(t38_state_t *s, int indicator);

int t38_gateway_process_rx_data(t38_state_t *s, int data_type, int field_type, const uint8_t *buf, int len);

int t38_gateway_process_rx_missing(t38_state_t *s, int seq_no);

int t38_terminal_process_rx_indicator(t38_state_t *s, int indicator);

int t38_terminal_process_rx_data(t38_state_t *s, int data_type, int field_type, const uint8_t *buf, int len);

int t38_terminal_process_rx_missing(t38_state_t *s, int seq_no);

int t38_send_timeout(t38_state_t *s);

int t38_terminal_init(t38_state_t *s,
                      int calling_party,
                      t38_tx_packet_handler_t *tx_packet_handler,
                      void *tx_packet_user_data);

int t38_gateway_init(t38_state_t *s,
                     int calling_party,
                     t38_tx_packet_handler_t *tx_packet_handler,
                     void *tx_packet_user_data);

int t38_rx_ifp_packet(t38_state_t *s, int seq_no, const uint8_t *buf, int len);

int t38_rx(t38_state_t *s, const int16_t *buf, int max_len);

int t38_tx(t38_state_t *s, int16_t *buf, int max_len);

#ifdef __cplusplus
}
#endif

#endif
/*- End of file ------------------------------------------------------------*/
