/*
 * SpanDSP - a series of DSP components for telephony
 *
 * t38.c - An implementation of a T.38 gateway, less the packet exchange part
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2005 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: t38_gateway.c,v 1.4 2005/10/13 17:22:51 steveu Exp $
 */

/*! \file */

#define	_ISOC9X_SOURCE	1
#define _ISOC99_SOURCE	1

#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <tgmath.h>
#include <math.h>
#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <sys/time.h>
#include <tiffio.h>

#include "spandsp.h"

#define ms_to_samples(t)    (((t)*SAMPLE_RATE)/1000)

enum
{
    T38_CED_TONE,
    T38_CNG_TONE,
    T38_V21_RX,
    T38_V27TER_RX,
    T38_V29_RX,
    T38_V17_RX
};

static int restart_rx_modem(t38_state_t *s);
static void add_to_non_ecm_tx_buffer(t38_state_t *s, const uint8_t *buf, int len);
static void t38_hdlc_tx(t38_state_t *s, const uint8_t *frame, int len);
static void t38_hdlc_tx_preamble(t38_state_t *s, int len);
static int t38_hdlc_tx_getbit(t38_state_t *t);
static void t38_hdlc_tx_init(t38_state_t *s);

static int send_data(t38_state_t *s, int data_type, int field_type, const uint8_t *msg, int msglen)
{
    uint8_t buf[100];
    int len;

    if ((len = t38_encode_data(s, buf, data_type, field_type, msg, msglen)) > 0)
        s->tx_packet_handler(s, s->tx_packet_user_data, buf, len);
    else
        span_log(&s->logging, SPAN_LOG_FLOW, "T.38 data len is %d\n", len);
    return 0;
}
/*- End of function --------------------------------------------------------*/

static int send_indicator(t38_state_t *s, int indicator)
{
    uint8_t buf[100];
    int len;

    if ((len = t38_encode_indicator(s, buf, indicator)) > 0)
    {
        s->tx_packet_handler(s, s->tx_packet_user_data, buf, len);
        s->current_tx_indicator = indicator;
    }
    else
    {
        span_log(&s->logging, SPAN_LOG_FLOW, "T.38 indicator len is %d\n", len);
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/

static int set_next_tx_type(t38_state_t *s)
{
    tone_gen_descriptor_t tone_desc;

    if (s->next_rx_indicator == s->current_rx_indicator)
    {
        span_log(&s->logging, SPAN_LOG_FLOW, "Changing to no signal - %d %d\n", s->next_rx_indicator, s->current_rx_indicator);
        s->silence = ms_to_samples(75);
        s->current_rx_indicator = T38_IND_NO_SIGNAL;
        return 0;
    }
    s->silence = -1;
    span_log(&s->logging, SPAN_LOG_FLOW, "Changing to %d\n", s->next_rx_indicator);

    switch (s->next_rx_indicator)
    {
    case T38_IND_NO_SIGNAL:
        /* Impose 75ms minimum on transmitted silence */
        s->silence = ms_to_samples(75);
        break;
    case T38_IND_CNG:
        /* 0.5s of 1100Hz + 3.0s of silence repeating */
        make_tone_gen_descriptor(&tone_desc,
                                 1100,
                                 -11,
                                 0,
                                 0,
                                 500,
                                 3000,
                                 0,
                                 0,
                                 TRUE);
        tone_gen_init(&(s->tone_gen), &tone_desc);
        break;
    case T38_IND_CED:
        make_tone_gen_descriptor(&tone_desc,
                                 2100,
                                 -11,
                                 0,
                                 0,
                                 2600,
                                 75,
                                 0,
                                 0,
                                 FALSE);
        tone_gen_init(&(s->tone_gen), &tone_desc);
        break;
    case T38_IND_V21_PREAMBLE:
        t38_hdlc_tx_init(s);
        t38_hdlc_tx_preamble(s, 32);
        s->data_final = FALSE;
        s->hdlc_len = 0;
        fsk_tx_init(&(s->v21tx), &preset_fsk_specs[FSK_V21CH2], (get_bit_func_t) t38_hdlc_tx_getbit, s);
        break;
    case T38_IND_V27TER_2400_TRAINING:
        v27ter_tx_restart(&(s->v27ter_tx), 2400, FALSE);
        break;
    case T38_IND_V27TER_4800_TRAINING:
        v27ter_tx_restart(&(s->v27ter_tx), 4800, FALSE);
        break;
    case T38_IND_V29_7200_TRAINING:
        v29_tx_restart(&(s->v29tx), 7200, FALSE);
        break;
    case T38_IND_V29_9600_TRAINING:
        v29_tx_restart(&(s->v29tx), 9600, FALSE);
        break;
#if defined(ENABLE_V17)
    case T38_IND_V17_7200_SHORT_TRAINING:
        v17_tx_restart(&(s->v17tx), 7200, TRUE, s->short_train);
        break;
    case T38_IND_V17_7200_LONG_TRAINING:
        v17_tx_restart(&(s->v17tx), 7200, FALSE, s->short_train);
        break;
    case T38_IND_V17_9600_SHORT_TRAINING:
        v17_tx_restart(&(s->v17tx), 9600, TRUE, s->short_train);
        break;
    case T38_IND_V17_9600_LONG_TRAINING:
        v17_tx_restart(&(s->v17tx), 9600, FALSE, s->short_train);
        break;
    case T38_IND_V17_12000_SHORT_TRAINING:
        v17_tx_restart(&(s->v17tx), 12000, TRUE, s->short_train);
        break;
    case T38_IND_V17_12000_LONG_TRAINING:
        v17_tx_restart(&(s->v17tx), 12000, FALSE, s->short_train);
        break;
    case T38_IND_V17_14400_SHORT_TRAINING:
        v17_tx_restart(&(s->v17tx), 14400, TRUE, s->short_train);
        break;
    case T38_IND_V17_14400_LONG_TRAINING:
        v17_tx_restart(&(s->v17tx), 14400, FALSE, s->short_train);
        break;
#endif
    case T38_IND_V8_ANSAM:
        break;
    case T38_IND_V8_SIGNAL:
        break;
    case T38_IND_V34_CNTL_CHANNEL_1200:
        break;
    case T38_IND_V34_PRI_CHANNEL:
        break;
    case T38_IND_V34_CC_RETRAIN:
        break;
    case T38_IND_V33_12000_TRAINING:
        break;
    case T38_IND_V33_14400_TRAINING:
        break;
    default:
        break;
    }
    s->tx_data_bytes = 0;
    s->bit_no = 0;
    s->current_byte = 0xFF;
    s->tx_in_bytes = 0;
    s->tx_latest_eol_byte = -1;
    s->bit_stream = 0;
    s->current_rx_indicator = s->next_rx_indicator;
    return 0;
}
/*- End of function --------------------------------------------------------*/

int t38_gateway_process_rx_missing(t38_state_t *s, int seq_no)
{
    span_log(&s->logging, SPAN_LOG_WARNING, "IFP seq no %d missing\n", seq_no);
    return 0;
}
/*- End of function --------------------------------------------------------*/

int t38_gateway_process_rx_indicator(t38_state_t *s, int indicator)
{
    span_log(&s->logging, SPAN_LOG_FLOW, "Rx indicator %s\n", t38_indicator(indicator));
    s->next_rx_indicator = indicator;

    if (s->silence == 0)
    {
        span_log(&s->logging, SPAN_LOG_FLOW, "Immediate change\n");
        set_next_tx_type(s);
    }
    else
    {
        span_log(&s->logging, SPAN_LOG_FLOW, "Delayed change - %d %d\n", s->current_rx_indicator, s->next_rx_indicator);
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/

static void pump_out_hdlc(t38_state_t *s, int previous)
{
    /* Don't start pumping data into the actual output stream until there is
       enough backlog to create some elasticity for jitter tolerance. */
    if (s->current_rx_indicator == T38_IND_V21_PREAMBLE  &&  s->hdlc_len > 5)
    {
        if (previous <= 5)
            previous = 0;
        t38_hdlc_tx(s, s->hdlc_buf + previous, s->hdlc_len - previous);
    }
}
/*- End of function --------------------------------------------------------*/

static void pump_out_final_hdlc(t38_state_t *s, int good_fcs)
{
    uint16_t crc;

    crc = crc_itu16_calc(s->hdlc_buf, s->hdlc_len);
    /* If the incoming CRC was wrong, corrupt the outgoing one */
    if (!good_fcs)
        crc = ~crc;
    s->hdlc_buf[s->hdlc_len++] = crc & 0xFF;
    s->hdlc_buf[s->hdlc_len++] = (crc >> 8) & 0xFF;
    if (s->current_rx_indicator == T38_IND_V21_PREAMBLE)
    {
        if (s->hdlc_len <= 5)
            t38_hdlc_tx(s, s->hdlc_buf, s->hdlc_len);
        else
            t38_hdlc_tx(s, s->hdlc_buf + s->hdlc_len - 2, 2);
        t38_hdlc_tx_preamble(s, 2);
    }
}
/*- End of function --------------------------------------------------------*/

int t38_gateway_process_rx_data(t38_state_t *s, int data_type, int field_type, const uint8_t *buf, int len)
{
    int i;
    int previous;

    span_log(&s->logging, SPAN_LOG_FLOW, "Rx data type %s/%s\n", t38_data_type(data_type), t38_field_type(field_type));
    switch (field_type)
    {
    case T38_FIELD_HDLC_DATA:
        previous = s->hdlc_len;
        for (i = 0;  i < len;  i++)
        {
            s->hdlc_buf[s->hdlc_len++] = bit_reverse(buf[i]);
            if (s->hdlc_len == 4)
            {
                /* Check if we need to corrupt this message */
                if (s->hdlc_buf[2] == T30_NSF
                    ||
                    s->hdlc_buf[2] == T30_NSC
                    ||
                    s->hdlc_buf[2] == T30_NSS)
                {
                    /* Corrupt the message, so it will be ignored */
                    s->hdlc_buf[3] = 0;
                }
            }
            else if (s->hdlc_len == 6)
            {
                if (s->hdlc_buf[2] == T30_DIS
                    ||
                    s->hdlc_buf[2] == T30_DTC
                    ||
                    s->hdlc_buf[2] == T30_DCS)
                {
                    /* We may need to adjust the capabilities, so they do not exceed our own */
                }
            }
        }
        pump_out_hdlc(s, previous);
        break;
    case T38_FIELD_HDLC_SIG_END:
        /* There should be no data */
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_SIG_END!\n");
        break;
    case T38_FIELD_HDLC_FCS_OK:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_FCS_OK!\n");
        span_log(&s->logging, SPAN_LOG_FLOW, "Type %s - CRC good\n", t30_frametype(s->hdlc_buf[2]));
        pump_out_final_hdlc(s, TRUE);
        s->hdlc_len = 0;
        break;
    case T38_FIELD_HDLC_FCS_BAD:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_FCS_BAD!\n");
        span_log(&s->logging, SPAN_LOG_FLOW, "Type %s - CRC bad\n", t30_frametype(s->hdlc_buf[2]));
        pump_out_final_hdlc(s, FALSE);
        s->hdlc_len = 0;
        break;
    case T38_FIELD_HDLC_FCS_OK_SIG_END:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_FCS_OK_SIG_END!\n");
        span_log(&s->logging, SPAN_LOG_FLOW, "Type %s - CRC OK, sig end\n", t30_frametype(s->hdlc_buf[2]));
#if 0
        switch (s->hdlc_buf[2])
        {
        case T30_DIS:
        case T30_DTC:
        case T30_DCS:
            t30_decode_dis_dtc_dcs(NULL, s->hdlc_buf, s->hdlc_len);
            break;
        default:
            break;
        }
#endif
        pump_out_final_hdlc(s, TRUE);
        s->data_final = TRUE;
        s->hdlc_len = 0;
        break;
    case T38_FIELD_HDLC_FCS_BAD_SIG_END:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_FCS_BAD_SIG_END!\n");
        span_log(&s->logging, SPAN_LOG_FLOW, "Type %s - CRC bad, sig end\n", t30_frametype(s->hdlc_buf[2]));
        pump_out_final_hdlc(s, FALSE);
        s->data_final = TRUE;
        s->hdlc_len = 0;
        break;
    case T38_FIELD_T4_NON_ECM_DATA:
        add_to_non_ecm_tx_buffer(s, buf, len);
        break;
    case T38_FIELD_T4_NON_ECM_SIG_END:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_NON_ECM_SIG_END!\n");
        /* Don't flow control the data any more */
        s->tx_latest_eol_byte = s->tx_in_bytes;
        s->data_final = TRUE;
        break;
    case T38_FIELD_CM_MESSAGE:
    case T38_FIELD_JM_MESSAGE:
    case T38_FIELD_CI_MESSAGE:
    case T38_FIELD_V34RATE:
    default:
        break;
    }

#if 0
    if (span_log_test(&s->logging, SPAN_LOG_FLOW))
    {
        int i;

        if (len > 0)
        {
            span_log(&s->logging, SPAN_LOG_FLOW, "Data: ");
            for (i = 0;  i < len;  i++)
                span_log(&s->logging, SPAN_LOG_FLOW | SPAN_LOG_SUPPRESS_LABELLING, " %02X", buf[i]);
        }
    }
    span_log(&s->logging, SPAN_LOG_FLOW | SPAN_LOG_SUPPRESS_LABELLING, "\n");
#endif
    return 0;
}
/*- End of function --------------------------------------------------------*/

#if defined(ENABLE_V17)
static void non_ecm_putbit_v17(void *user_data, int bit)
{
    t38_state_t *s;

    s = (t38_state_t *) user_data;
    if (bit < 0)
    {
        /* Special conditions */
        switch (bit)
        {
        case PUTBIT_TRAINING_FAILED:
            break;
        case PUTBIT_TRAINING_SUCCEEDED:
            /* The modem is now trained */
            s->rx_signal_present = TRUE;
            s->v21_rx_active = FALSE;
            s->v27ter_rx_active = FALSE;
            s->v29_rx_active = FALSE;
            send_indicator(s, T38_IND_V17_14400_LONG_TRAINING);
            s->current_tx_data = T38_DATA_V17_14400;
            s->since_last_tx_packet = 0;
            s->rx_data_bytes = 0;
            break;
        case PUTBIT_CARRIER_UP:
            break;
        case PUTBIT_CARRIER_DOWN:
            if (s->current_tx_indicator == T38_IND_V17_14400_LONG_TRAINING)
            {
                send_data(s, s->current_tx_data, T38_FIELD_T4_NON_ECM_SIG_END, "", 0);
                send_indicator(s, T38_IND_NO_SIGNAL);
                s->rx_signal_present = FALSE;
                restart_rx_modem(s);
            }
            break;
        default:
            span_log(&s->logging, SPAN_LOG_WARNING, "Eh!\n");
            break;
        }
        return;
    }
    s->current_byte = (s->current_byte >> 1) | (bit << 7);
    if (++s->bit_no >= 8)
    {
        s->rx_data[s->rx_data_bytes++] = s->current_byte;
        if (s->since_last_tx_packet >= 160)
        {
            send_data(s, s->current_tx_data, T38_FIELD_T4_NON_ECM_DATA, s->rx_data, s->rx_data_bytes);
            s->rx_data_bytes = 0;
            s->since_last_tx_packet = 0;
        }
        s->bit_no = 0;
        s->current_byte = 0;
    }
}
/*- End of function --------------------------------------------------------*/
#endif

static void non_ecm_putbit_v29(void *user_data, int bit)
{
    t38_state_t *s;

    s = (t38_state_t *) user_data;
    if (bit < 0)
    {
        /* Special conditions */
        switch (bit)
        {
        case PUTBIT_TRAINING_FAILED:
            break;
        case PUTBIT_TRAINING_SUCCEEDED:
            /* The modem is now trained */
            s->rx_signal_present = TRUE;
            s->v21_rx_active = FALSE;
            s->v17_rx_active = FALSE;
            s->v27ter_rx_active = FALSE;
            send_indicator(s, T38_IND_V29_9600_TRAINING);
            s->current_tx_data = T38_DATA_V29_9600;
            s->since_last_tx_packet = 0;
            s->rx_data_bytes = 0;
            break;
        case PUTBIT_CARRIER_UP:
            break;
        case PUTBIT_CARRIER_DOWN:
            if (s->current_tx_indicator == T38_IND_V29_9600_TRAINING)
            {
                send_data(s, s->current_tx_data, T38_FIELD_T4_NON_ECM_SIG_END, (uint8_t *) "", 0);
                send_indicator(s, T38_IND_NO_SIGNAL);
                s->rx_signal_present = FALSE;
                restart_rx_modem(s);
            }
            break;
        default:
            span_log(&s->logging, SPAN_LOG_WARNING, "Eh!\n");
            break;
        }
        return;
    }
    s->current_byte = (s->current_byte >> 1) | (bit << 7);
    if (++s->bit_no >= 8)
    {
        s->rx_data[s->rx_data_bytes++] = s->current_byte;
        if (s->since_last_tx_packet >= 160)
        {
            send_data(s, s->current_tx_data, T38_FIELD_T4_NON_ECM_DATA, s->rx_data, s->rx_data_bytes);
            s->rx_data_bytes = 0;
            s->since_last_tx_packet = 0;
        }
        s->bit_no = 0;
        s->current_byte = 0;
    }
}
/*- End of function --------------------------------------------------------*/

static void non_ecm_putbit_v27ter(void *user_data, int bit)
{
    t38_state_t *s;

    s = (t38_state_t *) user_data;
    if (bit < 0)
    {
        /* Special conditions */
        switch (bit)
        {
        case PUTBIT_TRAINING_FAILED:
            break;
        case PUTBIT_TRAINING_SUCCEEDED:
            /* The modem is now trained */
            s->rx_signal_present = TRUE;
            s->v21_rx_active = FALSE;
            s->v17_rx_active = FALSE;
            s->v29_rx_active = FALSE;
            send_indicator(s, T38_IND_V27TER_4800_TRAINING);
            s->current_tx_data = T38_DATA_V27TER_4800;
            s->since_last_tx_packet = 0;
            s->rx_data_bytes = 0;
            break;
        case PUTBIT_CARRIER_UP:
            break;
        case PUTBIT_CARRIER_DOWN:
            if (s->current_tx_indicator == T38_IND_V27TER_4800_TRAINING)
            {
                send_data(s, s->current_tx_data, T38_FIELD_T4_NON_ECM_SIG_END, (uint8_t *) "", 0);
                send_indicator(s, T38_IND_NO_SIGNAL);
                s->rx_signal_present = FALSE;
                restart_rx_modem(s);
            }
            break;
        default:
            span_log(&s->logging, SPAN_LOG_WARNING, "Eh!\n");
            break;
        }
        return;
    }
    s->current_byte = (s->current_byte >> 1) | (bit << 7);
    if (++s->bit_no >= 8)
    {
        s->rx_data[s->rx_data_bytes++] = s->current_byte;
        if (s->since_last_tx_packet >= 160)
        {
            send_data(s, s->current_tx_data, T38_FIELD_T4_NON_ECM_DATA, s->rx_data, s->rx_data_bytes);
            s->rx_data_bytes = 0;
            s->since_last_tx_packet = 0;
        }
        s->bit_no = 0;
        s->current_byte = 0;
    }
}
/*- End of function --------------------------------------------------------*/

static int non_ecm_getbit(void *user_data)
{
    t38_state_t *s;
    int bit;

    s = (t38_state_t *) user_data;
    if (s->bit_no <= 0)
    {
        /* We need another byte */
        if (s->tx_data_bytes != s->tx_latest_eol_byte)
        {
            s->current_byte = s->tx_data[s->tx_data_bytes];
            s->tx_data_bytes = (s->tx_data_bytes + 1) & (T38_TX_BUF_LEN - 1);
        }
        else
        {
            if (s->data_final)
            {
                s->data_final = FALSE;
                return 2;
            }
            if (s->tx_latest_eol_byte < 0)
            {
                /* Fill with 0xFF bytes. This is appropriate at the start of transmission,
                   as per T.30. */
                s->current_byte = 0xFF;
            }
            else
            {
                /* Hover at the last known EOL */
                s->current_byte = 0;
            }
        }
        s->bit_no = 8;
    }
    s->bit_no--;
    bit = s->current_byte & 1;
    s->current_byte >>= 1;
    return bit;
}
/*- End of function --------------------------------------------------------*/

static void add_to_non_ecm_tx_buffer(t38_state_t *s, const uint8_t *buf, int len)
{
    int i;
    int j;

    /* A rate adapting data stuffer for non-ECM image data */
    for (i = 0;  i < len;  i++)
    {
        s->tx_data[s->tx_in_bytes] = buf[i];
        s->tx_in_bytes = (s->tx_in_bytes + 1) & (T38_TX_BUF_LEN - 1);
        /* Check for EOLs, because at an EOL we can pause and pump out zeros while
           waiting for more incoming data. */
        s->bit_stream |= (buf[i] << 24);
        for (j = 0;  j < 8;  j++)
        {
            s->bit_stream >>= 1;
            if ((s->bit_stream & 0x00FFF000) == 0x00800000)
                s->tx_latest_eol_byte = (s->tx_in_bytes - 1) & (T38_TX_BUF_LEN - 1);
        }
    }
}
/*- End of function --------------------------------------------------------*/

static void t38_hdlc_tx_init(t38_state_t *s)
{
    s->hdlctx_num_bits = 0;
    s->hdlctx_len = 0;
    s->hdlctx_pos = 0;
    s->hdlctx_byte = 0;
    s->hdlctx_bits = 0;
    s->hdlctx_ones = 0;
    s->hdlctx_idle_byte = 0x7E;
    s->hdlctx_buffer[s->hdlctx_len++] = s->hdlctx_idle_byte;
    s->hdlctx_buffer[s->hdlctx_len++] = s->hdlctx_idle_byte;
    s->hdlctx_buffer[s->hdlctx_len] = s->hdlctx_idle_byte;
    s->hdlc_underflow_reported = FALSE;
}
/*- End of function --------------------------------------------------------*/

static void t38_hdlc_tx(t38_state_t *s, const uint8_t *frame, int len)
{
    int i;
    int byte_in_progress;

    while (len--)
    {
        byte_in_progress = *frame++;
        for (i = 0;  i < 8;  i++)
        {
            if (byte_in_progress & 0x01)
            {
                s->hdlctx_buffer[s->hdlctx_len] = (s->hdlctx_buffer[s->hdlctx_len] << 1) | 0x01;
                if (++s->hdlctx_ones >= 5)
                {
                    if (++s->hdlctx_num_bits == 8)
                    {
                        s->hdlctx_len++;
                        s->hdlctx_num_bits = 0;
                    }
                    s->hdlctx_buffer[s->hdlctx_len] <<= 1;
                    s->hdlctx_ones = 0;
                }
            }
            else
            {
                s->hdlctx_buffer[s->hdlctx_len] <<= 1;
                s->hdlctx_ones = 0;
            }
            if (++s->hdlctx_num_bits == 8)
            {
                s->hdlctx_len++;
                s->hdlctx_num_bits = 0;
            }
            byte_in_progress >>= 1;
        }
    }
    s->hdlctx_idle_byte = (0x7E7E >> s->hdlctx_num_bits) & 0xFF;
}
/*- End of function --------------------------------------------------------*/

static void t38_hdlc_tx_preamble(t38_state_t *s, int len)
{
    s->hdlctx_idle_byte = (0x7E7E >> s->hdlctx_num_bits) & 0xFF;
    if (len-- > 0)
    {
        s->hdlctx_buffer[s->hdlctx_len] <<= (8 - s->hdlctx_num_bits);
        s->hdlctx_buffer[s->hdlctx_len] |= (0x7E >> s->hdlctx_num_bits);
        s->hdlctx_len++;
    }
    while (len-- > 0)
        s->hdlctx_buffer[s->hdlctx_len++] = s->hdlctx_idle_byte;
    s->hdlctx_buffer[s->hdlctx_len] = 0x7E;
    s->hdlctx_ones = 0;
}
/*- End of function --------------------------------------------------------*/

static int t38_hdlc_tx_getbit(t38_state_t *s)
{
    int txbit;

    if (s->hdlctx_bits-- == 0)
    {
        if (s->hdlctx_len != s->hdlctx_pos)
        {
            s->hdlctx_byte = s->hdlctx_buffer[s->hdlctx_pos++];
            if (s->hdlctx_pos >= s->hdlctx_len)
            {
                s->hdlctx_pos = 0;
                s->hdlctx_len = 0;
                s->hdlctx_buffer[0] = 0x7E;
            }
            s->hdlc_underflow_reported = FALSE;
        }
        else
        {
            s->hdlctx_byte = s->hdlctx_idle_byte;
            if (!s->hdlc_underflow_reported)
            {
                if (s->data_final)
                {
                    set_next_tx_type(s);
                    s->data_final = FALSE;
                }
                s->hdlc_underflow_reported = TRUE;
            }
        }
        s->hdlctx_bits = 7;
    }
    txbit = (s->hdlctx_byte >> 7) & 0x01;
    s->hdlctx_byte <<= 1;
    return  txbit;
}
/*- End of function --------------------------------------------------------*/

static void t38_hdlc_rx_bit(hdlc_rx_state_t *s, int new_bit)
{
    t38_state_t *t;
    int final;
    uint8_t buf[2];

    t = (t38_state_t *) s->user_data;
    if (new_bit < 0)
    {
        /* Special conditions */
        switch (new_bit)
        {
        case PUTBIT_CARRIER_UP:
            /* Reset the HDLC receiver. */
            s->rx_state = FALSE;
            s->len = 0;
            s->num_bits = 0;
            t->rx_signal_present = TRUE;
            break;
        case PUTBIT_CARRIER_DOWN:
            if (t->current_tx_indicator == T38_IND_V21_PREAMBLE)
            {
                send_data(t, T38_DATA_V21, T38_FIELD_HDLC_SIG_END, (uint8_t *) "", 0);
                send_indicator(t, T38_IND_NO_SIGNAL);
                t->rx_signal_present = FALSE;
                restart_rx_modem(t);
            }
            t->rx_flag_count = 0;
            break;
        default:
            span_log(&t->logging, SPAN_LOG_WARNING, "Unexpected HDLC special bit - %d!\n", new_bit);
            break;
        }
        return;
    }
    s->bit_buf |= (new_bit & 1);
    if ((s->bit_buf & 0x7F) == 0x7E)
    {
        if ((s->bit_buf & 0x80))
        {
            /* Hit HDLC abort */
            s->rx_aborts++;
            s->rx_state = FALSE;
        }
        else
        {
            /* Hit HDLC flag */
            if (s->rx_state)
            {
                if (t->rx_flag_count < 4  &&  ++t->rx_flag_count >= 4)
                {
                    send_indicator(t, T38_IND_V21_PREAMBLE);
                    t->v17_rx_active = FALSE;
                    t->v27ter_rx_active = FALSE;
                    t->v29_rx_active = FALSE;
                    if (t->current_rx_indicator == T38_IND_CNG)
                    {
                        t->next_rx_indicator = T38_IND_NO_SIGNAL;
                        set_next_tx_type(t);
                    }
                }
                if (s->len)
                {
                    if (s->len >= 2)
                    {
                        if (crc_itu16_check(s->buffer, s->len))
                        {
                            s->rx_frames++;
                            s->rx_bytes += s->len - 2;
                            t->rx_data_bytes = 0;
                            final = (s->len - 2 >= 2  &&  s->buffer[0] == 0xFF  &&  s->buffer[1] == 0x13);
                            span_log(&t->logging, SPAN_LOG_FLOW, "E Type %s\n", t30_frametype(s->buffer[2]));
                            send_data(t, T38_DATA_V21, (final)  ?  T38_FIELD_HDLC_FCS_OK_SIG_END  :  T38_FIELD_HDLC_FCS_OK, (uint8_t *) "", 0);
                        }
                        else
                        {
                            s->rx_crc_errors++;
                            t->rx_data_bytes = 0;
                            final = (s->len - 2 >= 2  &&  s->buffer[0] == 0xFF  &&  s->buffer[1] == 0x13);
                            span_log(&t->logging, SPAN_LOG_FLOW, "F Type %s\n", t30_frametype(s->buffer[2]));
                            send_data(t, T38_DATA_V21, (final)  ?  T38_FIELD_HDLC_FCS_BAD_SIG_END  :  T38_FIELD_HDLC_FCS_BAD, (uint8_t *) "", 0);
                        }
                    }
                    else
                    {
                        /* Frame too short */
                        s->rx_length_errors++;
                    }
                }
            }
            else
            {
                s->rx_state = TRUE;
            }
            s->len = 0;
            s->num_bits = 0;
            t->corrupt_frame = FALSE;
        }
    }
    else
    {
        if (s->rx_state  &&  (s->bit_buf & 0x3F) != 0x3E)
        {
            s->byte_in_progress >>= 1;
            s->byte_in_progress |= ((s->bit_buf & 0x01) << 7);
            if (++s->num_bits == 8)
            {
                if (s->len >= sizeof(s->buffer))
                {
                    /* Frame too long */
                    s->rx_length_errors++;
                    s->rx_state = FALSE;
                }
                else
                {
                    s->buffer[s->len] = s->byte_in_progress;
                    if (s->len >= 2)
                    {
                        /* Make the transmission lag by two bytes, so we do not send the CRC, and
                           do not report the CRC result too late. */
                        buf[0] = (t->corrupt_frame)  ?  0  :  bit_reverse(s->buffer[s->len - 2]);
                        if (s->len == 4)
                        {
                            /* We have the frame type. Check if we need to corrupt this message */
                            if (s->buffer[2] == T30_NSF
                                ||
                                s->buffer[2] == T30_NSC
                                ||
                                s->buffer[2] == T30_NSS)
                            {
                                t->corrupt_frame = TRUE;
                            }
                        }
                        send_data(t, T38_DATA_V21, T38_FIELD_HDLC_DATA, buf, 1);
                    }
                    s->len++;
                    s->num_bits = 0;
                }
            }
        }
    }
    s->bit_buf <<= 1;
}
/*- End of function --------------------------------------------------------*/

static int restart_rx_modem(t38_state_t *s)
{
    span_log(&s->logging, SPAN_LOG_FLOW, "Restart modem\n");
    /* Turn all the modems on, and let them sort out what is really happening */
    hdlc_rx_init(&(s->hdlcrx), FALSE, TRUE, 5, NULL, s);

    s->rx_signal_present = FALSE;
    s->rx_flag_count = 0;
    s->data_final = FALSE;
    fsk_rx_init(&(s->v21rx), &preset_fsk_specs[FSK_V21CH2], TRUE, (put_bit_func_t) t38_hdlc_rx_bit, &(s->hdlcrx));
    s->v21_rx_active = TRUE;
#if defined(ENABLE_V17)
    v17_rx_restart(&(s->v17rx), s->bit_rate, s->short_train);
    s->v17_rx_active = TRUE;
#endif
    v27ter_rx_restart(&(s->v27ter_rx), s->bit_rate);
    s->v27ter_rx_active = TRUE;
    v29_rx_restart(&(s->v29rx), s->bit_rate);
    s->v29_rx_active = TRUE;
    return 0;
}
/*- End of function --------------------------------------------------------*/

int t38_rx(t38_state_t *s, const int16_t *buf, int len)
{
    s->since_last_tx_packet += len;
    if (s->v21_rx_active)
        fsk_rx(&(s->v21rx), buf, len);
#if defined(ENABLE_V17)
    if (s->v17_rx_active)
        v17_rx(&(s->v17rx), buf, len);
#endif
    if (s->v27ter_rx_active)
        v27ter_rx(&(s->v27ter_rx), buf, len);
    if (s->v29_rx_active)
        v29_rx(&(s->v29rx), buf, len);
    return  0;
}
/*- End of function --------------------------------------------------------*/

int t38_tx(t38_state_t *s, int16_t *buf, int max_len)
{
    int len;

    len = 0;
    if (s->silence == 0)
        return len;
    switch (s->current_rx_indicator)
    {
    case T38_IND_NO_SIGNAL:
        if (s->silence > 0)
        {
            len = max_len;
            if (len > s->silence)
                len = s->silence;
            s->silence -= len;
            memset(buf, 0, len*sizeof(int16_t));
        }
        break;
    case T38_IND_CED:
        if ((len = tone_gen(&(s->tone_gen), buf + len, max_len)) <= 0)
        {
            s->silence = 0;
            set_next_tx_type(s);
        }
        break;
    case T38_IND_CNG:
        len = tone_gen(&(s->tone_gen), buf + len, max_len);
        break;
    case T38_IND_V21_PREAMBLE:
        len = fsk_tx(&(s->v21tx), buf + len, max_len);
        break;
#if defined(ENABLE_V17)
    case T38_IND_V17_7200_LONG_TRAINING:
    case T38_IND_V17_9600_LONG_TRAINING:
    case T38_IND_V17_12000_LONG_TRAINING:
    case T38_IND_V17_14400_LONG_TRAINING:
    case T38_IND_V17_7200_SHORT_TRAINING:
    case T38_IND_V17_9600_SHORT_TRAINING:
    case T38_IND_V17_12000_SHORT_TRAINING:
    case T38_IND_V17_14400_SHORT_TRAINING:
        if ((len = v17_tx(&(s->v17tx), buf + len, max_len)) <= 0)
        {
            s->silence = 0;
            set_next_tx_type(s);
        }
        break;
#endif
    case T38_IND_V27TER_2400_TRAINING:
    case T38_IND_V27TER_4800_TRAINING:
        if ((len = v27ter_tx(&(s->v27ter_tx), buf + len, max_len)) <= 0)
        {
            s->silence = 0;
            set_next_tx_type(s);
        }
        break;
    case T38_IND_V29_7200_TRAINING:
    case T38_IND_V29_9600_TRAINING:
        if ((len = v29_tx(&(s->v29tx), buf + len, max_len)) <= 0)
        {
            s->silence = 0;
            set_next_tx_type(s);
        }
        break;
    }
    return len;
}
/*- End of function --------------------------------------------------------*/

int t38_gateway_init(t38_state_t *s,
                     int calling_party,
                     t38_tx_packet_handler_t *tx_packet_handler,
                     void *tx_packet_user_data)
{
    memset(s, 0, sizeof(*s));
    if (tx_packet_handler == NULL)
        return -1;

    s->gateway = TRUE;
#if defined(ENABLE_V17)
    v17_rx_init(&(s->v17rx), 14400, non_ecm_putbit_v17, s);
    v17_tx_init(&(s->v17tx), 14400, FALSE, non_ecm_getbit, s);
#endif
    v29_rx_init(&(s->v29rx), 9600, non_ecm_putbit_v29, s);
    v29_tx_init(&(s->v29tx), 9600, FALSE, non_ecm_getbit, s);
    v27ter_rx_init(&(s->v27ter_rx), 4800, non_ecm_putbit_v27ter, s);
    v27ter_tx_init(&(s->v27ter_tx), 4800, FALSE, non_ecm_getbit, s);
    s->rx_signal_present = FALSE;

    s->data_rate_management_method = 2;
    s->data_transport_protocol = T38_TRANSPORT_UDPTL;
    s->fill_bit_removal = FALSE;
    s->mmr_transcoding = FALSE;
    s->jbig_transcoding = FALSE;
    s->max_buffer_size = 400;
    s->max_datagram_size = 100;
    s->error_correction_scheme = -1;
    s->error_correction_span = 3;
    s->error_correction_entries = 2;
    s->t38_version = 1;

    gettimeofday(&s->last_time, NULL);

    s->silence = 0;
    restart_rx_modem(s);
    s->tx_packet_handler = tx_packet_handler;
    s->tx_packet_user_data = tx_packet_user_data;
    return 0;
}
/*- End of function --------------------------------------------------------*/
/*- End of file ------------------------------------------------------------*/
