/*
 * SpanDSP - a series of DSP components for telephony
 *
 * t38.c - An implementation of a T.38 terminal, less the packet exchange part
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2005 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: t38_terminal.c,v 1.3 2005/10/13 17:22:51 steveu Exp $
 */

/*! \file */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <sys/time.h>
#include <tiffio.h>

#include "spandsp.h"

#define ms_to_samples(t)            (((t)*SAMPLE_RATE)/1000)

#define MS_PER_TX_CHUNK             30

enum
{
    T38_CED_TONE,
    T38_CNG_TONE,
    T38_V21_RX,
    T38_V27TER_RX,
    T38_V29_RX,
    T38_V17_RX
};

static int send_data(t38_state_t *s, int data_type, int field_type, const uint8_t *msg, int msglen)
{
    uint8_t buf[100];
    int len;

    if ((len = t38_encode_data(s, buf, data_type, field_type, msg, msglen)) > 0)
        s->tx_packet_handler(s, s->tx_packet_user_data, buf, len);
    else
        span_log(&s->logging, SPAN_LOG_FLOW, "T.38 data len is %d\n", len);
    return 0;
}
/*- End of function --------------------------------------------------------*/

static int send_indicator(t38_state_t *s, int indicator)
{
    uint8_t buf[100];
    int len;

    if ((len = t38_encode_indicator(s, buf, indicator)) > 0)
    {
        s->tx_packet_handler(s, s->tx_packet_user_data, buf, len);
        s->current_tx_indicator = indicator;
    }
    else
    {
        span_log(&s->logging, SPAN_LOG_FLOW, "T.38 indicator len is %d\n", len);
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/

static int get_non_ecm_image_chunk(t38_state_t *s, uint8_t *buf, int len)
{
    int i;
    int j;
    int bit;
    int byte;

    for (i = 0;  i < len;  i++)
    {
        byte = 0;
        for (j = 0;  j < 8;  j++)
        {
            bit = t30_non_ecm_getbit(&s->t30_state);
            byte = (byte >> 1) | ((bit << 7) & 0x80);
            if (bit & 2)
            {
                byte >>= (7 - j);
                buf[i++] = byte;
                return -i;
            }
        }
        buf[i] = byte;
    }
    return i;
}
/*- End of function --------------------------------------------------------*/

static void put_non_ecm_image_chunk(t38_state_t *s, const uint8_t *buf, int len)
{
    int i;
    int j;
    int byte;

    for (i = 0;  i < len;  i++)
    {
        byte = buf[i];
        for (j = 0;  j < 8;  j++)
        {
            t30_non_ecm_putbit(&(s->t30_state), (byte & 1));
            byte >>= 1;
        }
    }
}
/*- End of function --------------------------------------------------------*/

static void set_next_time(struct timeval *then, struct timeval *now, int step)
{
    /* This logic assumes adding step to tv_usec cannot overflow */
    if (now)
    {
        then->tv_sec = now->tv_sec;
        then->tv_usec = now->tv_usec + step;
        if (then->tv_usec >= 1000000)
        {
            then->tv_sec = now->tv_sec + then->tv_usec/1000000;
            then->tv_usec %= 1000000;
        }
    }
    else
    {
        then->tv_usec += step;
        if (then->tv_usec >= 1000000)
        {
            then->tv_sec += then->tv_usec/1000000;
            then->tv_usec %= 1000000;
        }
    }
}
/*- End of function --------------------------------------------------------*/

int t38_terminal_process_rx_missing(t38_state_t *s, int seq_no)
{
    span_log(&s->logging, SPAN_LOG_FLOW, "IFP seq no %d missing\n", seq_no);
    return 0;
}
/*- End of function --------------------------------------------------------*/

int t38_terminal_process_rx_indicator(t38_state_t *s, int indicator)
{
    span_log(&s->logging, SPAN_LOG_FLOW, "Rx indicator %s\n", t38_indicator(indicator));
    switch (indicator)
    {
    case T38_IND_NO_SIGNAL:
        if (s->current_rx_indicator == T38_IND_V21_PREAMBLE
            &&
            (s->current_rx_type == T30_MODEM_V21  ||  s->current_rx_type == T30_MODEM_CNG))
        {
            t30_hdlc_accept(&(s->t30_state), TRUE, NULL, PUTBIT_CARRIER_DOWN);
        }
        break;
    case T38_IND_CNG:
        break;
    case T38_IND_CED:
        break;
    case T38_IND_V21_PREAMBLE:
        if (s->current_rx_type == T30_MODEM_V21)
        {
            t30_hdlc_accept(&(s->t30_state), TRUE, NULL, PUTBIT_CARRIER_UP);
            t30_hdlc_accept(&(s->t30_state), TRUE, NULL, PUTBIT_FRAMING_OK);
        }
        s->data_final = FALSE;
        s->hdlc_len = 0;
        s->tx_data_bytes = 0;
        break;
    case T38_IND_V27TER_2400_TRAINING:
        break;
    case T38_IND_V27TER_4800_TRAINING:
        break;
    case T38_IND_V29_7200_TRAINING:
        break;
    case T38_IND_V29_9600_TRAINING:
        break;
    case T38_IND_V17_7200_SHORT_TRAINING:
        break;
    case T38_IND_V17_7200_LONG_TRAINING:
        break;
    case T38_IND_V17_9600_SHORT_TRAINING:
        break;
    case T38_IND_V17_9600_LONG_TRAINING:
        break;
    case T38_IND_V17_12000_SHORT_TRAINING:
        break;
    case T38_IND_V17_12000_LONG_TRAINING:
        break;
    case T38_IND_V17_14400_SHORT_TRAINING:
        break;
    case T38_IND_V17_14400_LONG_TRAINING:
        break;
    case T38_IND_V8_ANSAM:
        break;
    case T38_IND_V8_SIGNAL:
        break;
    case T38_IND_V34_CNTL_CHANNEL_1200:
        break;
    case T38_IND_V34_PRI_CHANNEL:
        break;
    case T38_IND_V34_CC_RETRAIN:
        break;
    case T38_IND_V33_12000_TRAINING:
        break;
    case T38_IND_V33_14400_TRAINING:
        break;
    default:
        break;
    }
    s->current_rx_indicator = indicator;
    s->tx_data_bytes = 0;
    return 0;
}
/*- End of function --------------------------------------------------------*/

int t38_terminal_process_rx_data(t38_state_t *s, int data_type, int field_type, const uint8_t *buf, int len)
{
    int i;

    span_log(&s->logging, SPAN_LOG_FLOW, "Rx data type %s/%s\n", t38_data_type(data_type), t38_field_type(field_type));
    switch (data_type)
    {
    case T38_DATA_V21:
    case T38_DATA_V27TER_2400:
    case T38_DATA_V27TER_4800:
    case T38_DATA_V29_7200:
    case T38_DATA_V29_9600:
    case T38_DATA_V17_7200:
    case T38_DATA_V17_9600:
    case T38_DATA_V17_12000:
    case T38_DATA_V17_14400:
    case T38_DATA_V8:
    case T38_DATA_V34_PRI_RATE:
    case T38_DATA_V34_CC_1200:
    case T38_DATA_V34_PRI_CH:
    case T38_DATA_V33_12000:
    case T38_DATA_V33_14400:
    default:
        break;
    }

    switch (field_type)
    {
    case T38_FIELD_HDLC_DATA:
        for (i = 0;  i < len;  i++)
            s->tx_data[s->tx_data_bytes++] = bit_reverse(buf[i]);
        break;
    case T38_FIELD_HDLC_SIG_END:
        /* There should be no data */
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_SIG_END!\n");
        /* This is only expected when the HDLC signal drops unexpectedly.
           If it drops normally, following the end of a final frame, this
           message is not expected. */
        if (s->current_rx_type == T30_MODEM_V21)
            t30_hdlc_accept(&(s->t30_state), TRUE, NULL, PUTBIT_CARRIER_DOWN);
         s->tx_data_bytes = 0;
       break;
    case T38_FIELD_HDLC_FCS_OK:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_FCS_OK!\n");
        span_log(&s->logging, SPAN_LOG_FLOW, "Type %s - CRC OK\n", t30_frametype(s->tx_data[2]));
        if (s->current_rx_type == T30_MODEM_V21)
            t30_hdlc_accept(&(s->t30_state), TRUE, s->tx_data, s->tx_data_bytes);
        s->tx_data_bytes = 0;
        break;
    case T38_FIELD_HDLC_FCS_BAD:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_FCS_BAD!\n");
        span_log(&s->logging, SPAN_LOG_FLOW, "Type %s - CRC bad\n", t30_frametype(s->tx_data[2]));
        if (s->current_rx_type == T30_MODEM_V21)
            t30_hdlc_accept(&(s->t30_state), FALSE, s->tx_data, s->tx_data_bytes);
        s->tx_data_bytes = 0;
        break;
    case T38_FIELD_HDLC_FCS_OK_SIG_END:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_FCS_OK_SIG_END!\n");
        span_log(&s->logging, SPAN_LOG_FLOW, "Type %s - CRC OK, sig end\n", t30_frametype(s->tx_data[2]));
        if (s->current_rx_type == T30_MODEM_V21)
        {
            t30_hdlc_accept(&(s->t30_state), TRUE, s->tx_data, s->tx_data_bytes);
            t30_hdlc_accept(&(s->t30_state), TRUE, NULL, PUTBIT_CARRIER_DOWN);
        }
        s->tx_data_bytes = 0;
        break;
    case T38_FIELD_HDLC_FCS_BAD_SIG_END:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_HDLC_FCS_BAD_SIG_END!\n");
        span_log(&s->logging, SPAN_LOG_FLOW, "Type %s - CRC bad, sig end\n", t30_frametype(s->tx_data[2]));
        if (s->current_rx_type == T30_MODEM_V21)
        {
            t30_hdlc_accept(&(s->t30_state), FALSE, s->tx_data, s->tx_data_bytes);
            t30_hdlc_accept(&(s->t30_state), TRUE, NULL, PUTBIT_CARRIER_DOWN);
        }
        s->tx_data_bytes = 0;
        break;
    case T38_FIELD_T4_NON_ECM_DATA:
        if (!s->rx_signal_present)
        {
            t30_non_ecm_putbit(&(s->t30_state), PUTBIT_TRAINING_SUCCEEDED);
            s->rx_signal_present = TRUE;
        }
        put_non_ecm_image_chunk(s, buf, len);
        break;
    case T38_FIELD_T4_NON_ECM_SIG_END:
        if (len > 0)
            span_log(&s->logging, SPAN_LOG_WARNING, "There is data in a T38_FIELD_NON_ECM_SIG_END!\n");
        t30_non_ecm_putbit(&(s->t30_state), PUTBIT_CARRIER_DOWN);
        s->rx_signal_present = FALSE;
        break;
    case T38_FIELD_CM_MESSAGE:
    case T38_FIELD_JM_MESSAGE:
    case T38_FIELD_CI_MESSAGE:
    case T38_FIELD_V34RATE:
    default:
        break;
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/

void t38_send_hdlc(void *user_data, const uint8_t *msg, int len)
{
    t38_state_t *s;

    s = (t38_state_t *) user_data;

    if (s->hdlc_len == 0)
    {
        memcpy(s->hdlc_buf, msg, len);
        s->hdlc_len = len;
        s->hdlc_ptr = 0;
    }
    else if (s->hdlc_len2 == 0)
    {
        memcpy(s->hdlc_buf2, msg, len);
        s->hdlc_len2 = len;
    }
    else if (s->hdlc_len3 == 0)
    {
        memcpy(s->hdlc_buf3, msg, len);
        s->hdlc_len3 = len;
    }
}
/*- End of function --------------------------------------------------------*/

int t38_send_timeout(t38_state_t *s)
{
    int len;
    uint8_t buf[100];
    struct timeval now;
    int diff;


    gettimeofday(&now, NULL);
    
    diff = (now.tv_sec - s->last_time.tv_sec)*1000000 + (now.tv_usec - s->last_time.tv_usec);
    if (diff == 0)
        return 0;
    
    t30_timer_update(&s->t30_state, diff/125);
    s->last_time = now;
    
    if (s->timed_step == 0)
        return 0;
    if (now.tv_sec < s->next_send_time.tv_sec)
        return 0;
    if (now.tv_sec == s->next_send_time.tv_sec  &&  now.tv_usec < s->next_send_time.tv_usec)
        return 0;
    
    switch (s->timed_step)
    {
    case 1:
        /* Create a 75ms silence */
        send_indicator(s, T38_IND_NO_SIGNAL);
        s->timed_step = 2;
        set_next_time(&s->next_send_time, NULL, 75000);
        break;
    case 2:
        /* Switch on a fast modem, and give the training time to complete */
        send_indicator(s, s->next_tx_indicator);
        s->timed_step = 3;
        set_next_time(&s->next_send_time, NULL, 253000);
        break;
    case 3:
        /* Send a chunk of non-ECM image data */
        len = get_non_ecm_image_chunk(s, buf, s->octets_per_non_ecm_packet);
        send_data(s, s->current_tx_data, T38_FIELD_T4_NON_ECM_DATA, buf, abs(len));
        if (len > 0)
        {
            set_next_time(&s->next_send_time, NULL, MS_PER_TX_CHUNK*1000);
        }
        else
        {
            send_data(s, s->current_tx_data, T38_FIELD_T4_NON_ECM_SIG_END, NULL, 0);
            send_indicator(s, T38_IND_NO_SIGNAL);
            s->timed_step = 0;
            t30_send_complete(&(s->t30_state));
        }
        break;
    case 4:
        /* Send HDLC preambling */
        send_indicator(s, T38_IND_V21_PREAMBLE);
        set_next_time(&s->next_send_time, NULL, 1000000);
        s->timed_step = 5;
        break;
    case 5:
        /* Send HDLC octet */
        buf[0] = bit_reverse(s->hdlc_buf[s->hdlc_ptr++]);
        send_data(s, T38_DATA_V21, T38_FIELD_HDLC_DATA, buf, 1);
        if (s->hdlc_ptr >= s->hdlc_len)
            s->timed_step = 6;
        set_next_time(&s->next_send_time, NULL, MS_PER_TX_CHUNK*1000);
        break;
    case 6:
        /* End of HDLC frame */
        s->hdlc_ptr = 0;
        if (s->hdlc_len2)
        {
            /* Another frame follows */
            memcpy(s->hdlc_buf, s->hdlc_buf2, s->hdlc_len2);
            memcpy(s->hdlc_buf2, s->hdlc_buf3, s->hdlc_len3);
            s->hdlc_len = s->hdlc_len2;
            s->hdlc_len2 = s->hdlc_len3;
            s->hdlc_len3 = 0;
            send_data(s, T38_DATA_V21, T38_FIELD_HDLC_FCS_OK, NULL, 0);
            s->timed_step = 5;
        }
        else
        {
            s->hdlc_len = 0;
            send_data(s, T38_DATA_V21, T38_FIELD_HDLC_FCS_OK_SIG_END, NULL, 0);
            s->timed_step = 7;
            t30_send_complete(&s->t30_state);
        }
        set_next_time(&s->next_send_time, NULL, MS_PER_TX_CHUNK*1000);
        break;
    case 7:
        /* End of HDLC transmission */
        send_data(s, T38_DATA_V21, T38_FIELD_HDLC_SIG_END, NULL, 0);
        s->timed_step = 0;
        break;
    case 8:
        s->timed_step = 0;
        t30_send_complete(&s->t30_state);
        break;
    case 9:
        s->timed_step = 0;
        t30_send_complete(&s->t30_state);
        break;
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/

void t38_set_rx_type(void *user_data, int type, int short_train)
{
    t38_state_t *s;

    s = (t38_state_t *) user_data;
    span_log(&s->logging, SPAN_LOG_FLOW, "Set rx type %d\n", type);
    s->current_rx_type = type;
}
/*- End of function --------------------------------------------------------*/

void t38_set_tx_type(void *user_data, int type, int short_train)
{
    t38_state_t *s;
    struct timeval now;

    s = (t38_state_t *) user_data;
    span_log(&s->logging, SPAN_LOG_FLOW, "Set tx type %d\n", type);
    if (s->current_tx_type == type)
        return;
    gettimeofday(&now, NULL);

    switch (type)
    {
    case T30_MODEM_PAUSE:
        set_next_time(&s->next_send_time, &now, short_train*1000);
        s->timed_step = 8;
        send_indicator(s, T38_IND_NO_SIGNAL);
        break;
    case T30_MODEM_CED:
        set_next_time(&s->next_send_time, &now, 3000000);
        s->timed_step = 9;
        send_indicator(s, T38_IND_CED);
        break;
    case T30_MODEM_CNG:
        send_indicator(s, T38_IND_CNG);
        break;
    case T30_MODEM_V21:
        if (s->current_tx_type > T30_MODEM_V21)
        {
            /* Pause before switching from phase C, as per T.30. If we omit this, the receiver
               might not see the carrier fall between the high speed and low speed sections. */
            send_indicator(s, T38_IND_NO_SIGNAL);
            set_next_time(&s->next_send_time, &now, 75000);
            s->timed_step = 4;
        }
        else
        {
            send_indicator(s, T38_IND_V21_PREAMBLE);
            set_next_time(&s->next_send_time, &now, 1000000);
            s->timed_step = 5;
        }
        break;
    case T30_MODEM_V27TER_2400:
        s->octets_per_non_ecm_packet = MS_PER_TX_CHUNK*2400/8000;
        s->next_tx_indicator = T38_IND_V27TER_2400_TRAINING;
        s->current_tx_data = T38_DATA_V27TER_2400;
        set_next_time(&s->next_send_time, &now, MS_PER_TX_CHUNK*1000);
        s->timed_step = 1;
        break;
    case T30_MODEM_V27TER_4800:
        s->octets_per_non_ecm_packet = MS_PER_TX_CHUNK*4800/8000;
        s->next_tx_indicator = T38_IND_V27TER_4800_TRAINING;
        s->current_tx_data = T38_DATA_V27TER_4800;
        set_next_time(&s->next_send_time, &now, MS_PER_TX_CHUNK*1000);
        s->timed_step = 1;
        break;
    case T30_MODEM_V29_7200:
        s->octets_per_non_ecm_packet = MS_PER_TX_CHUNK*7200/8000;
        s->next_tx_indicator = T38_IND_V29_7200_TRAINING;
        s->current_tx_data = T38_DATA_V29_7200;
        set_next_time(&s->next_send_time, &now, MS_PER_TX_CHUNK*1000);
        s->timed_step = 1;
        break;
    case T30_MODEM_V29_9600:
        s->octets_per_non_ecm_packet = MS_PER_TX_CHUNK*9600/8000;
        s->next_tx_indicator = T38_IND_V29_9600_TRAINING;
        s->current_tx_data = T38_DATA_V29_9600;
        set_next_time(&s->next_send_time, &now, MS_PER_TX_CHUNK*1000);
        s->timed_step = 1;
        break;
    case T30_MODEM_V17_7200:
        s->octets_per_non_ecm_packet = MS_PER_TX_CHUNK*7200/8000;
        s->next_tx_indicator = (short_train)  ?  T38_IND_V17_7200_SHORT_TRAINING  :  T38_IND_V17_7200_LONG_TRAINING;
        s->current_tx_data = T38_DATA_V17_7200;
        set_next_time(&s->next_send_time, &now, MS_PER_TX_CHUNK*1000);
        s->timed_step = 1;
        break;
    case T30_MODEM_V17_9600:
        s->octets_per_non_ecm_packet = MS_PER_TX_CHUNK*9600/8000;
        s->next_tx_indicator = (short_train)  ?  T38_IND_V17_9600_SHORT_TRAINING  :  T38_IND_V17_9600_LONG_TRAINING;
        s->current_tx_data = T38_DATA_V17_9600;
        set_next_time(&s->next_send_time, &now, MS_PER_TX_CHUNK*1000);
        s->timed_step = 1;
        break;
    case T30_MODEM_V17_12000:
        s->octets_per_non_ecm_packet = MS_PER_TX_CHUNK*12000/8000;
        s->next_tx_indicator = (short_train)  ?  T38_IND_V17_12000_SHORT_TRAINING  :  T38_IND_V17_12000_LONG_TRAINING;
        s->current_tx_data = T38_DATA_V17_12000;
        set_next_time(&s->next_send_time, &now, MS_PER_TX_CHUNK*1000);
        s->timed_step = 1;
        break;
    case T30_MODEM_V17_14400:
        s->octets_per_non_ecm_packet = MS_PER_TX_CHUNK*14400/8000;
        s->next_tx_indicator = (short_train)  ?  T38_IND_V17_14400_SHORT_TRAINING  :  T38_IND_V17_14400_LONG_TRAINING;
        s->current_tx_data = T38_DATA_V17_14400;
        set_next_time(&s->next_send_time, &now, MS_PER_TX_CHUNK*1000);
        s->timed_step = 1;
        break;
    }
    s->current_tx_type = type;
}
/*- End of function --------------------------------------------------------*/

int t38_terminal_init(t38_state_t *s,
                      int calling_party,
                      t38_tx_packet_handler_t *tx_packet_handler,
                      void *tx_packet_user_data)
{
    memset(s, 0, sizeof(*s));
    if (tx_packet_handler == NULL)
        return -1;
    s->rx_signal_present = FALSE;

    s->gateway = FALSE;
    s->timed_step = 0;
    s->hdlc_ptr = 0;
    s->tx_packet_handler = tx_packet_handler;
    s->tx_packet_user_data = tx_packet_user_data;

    s->data_rate_management_method = 2;
    s->data_transport_protocol = T38_TRANSPORT_UDPTL;
    s->fill_bit_removal = FALSE;
    s->mmr_transcoding = FALSE;
    s->jbig_transcoding = FALSE;
    s->max_buffer_size = 400;
    s->max_datagram_size = 100;
    s->error_correction_scheme = -1;
    s->error_correction_span = 3;
    s->error_correction_entries = 2;
    s->t38_version = 1;

    t30_init(&(s->t30_state), calling_party, NULL);
    s->t30_state.set_rx_type_handler = t38_set_rx_type;
    s->t30_state.set_rx_type_user_data = (void *) s;
    s->t30_state.set_tx_type_handler = t38_set_tx_type;
    s->t30_state.set_tx_type_user_data = (void *) s;
    s->t30_state.send_hdlc_handler = t38_send_hdlc;
    s->t30_state.send_hdlc_user_data = (void *) s;

    gettimeofday(&s->last_time, NULL);

    t30_restart(&s->t30_state);
    return 0;
}
/*- End of function --------------------------------------------------------*/
/*- End of file ------------------------------------------------------------*/
